#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Action and ActionManager."""

from datetime import timedelta
from typing import TYPE_CHECKING, Any, Dict, Generator, Optional

from graviti.exception import ResourceNameError
from graviti.manager.common import LIMIT
from graviti.manager.lazy import LazyPagingList
from graviti.openapi import create_action, delete_action, get_action, list_actions, update_action
from graviti.openapi.action import create_action_run, get_action_run, list_action_runs
from graviti.utility import ReprMixin, check_type
from graviti.utility.common import convert_iso_to_datetime
from graviti.utility.typing import SortParam

if TYPE_CHECKING:
    from graviti.manager.dataset import Dataset


class Action(ReprMixin):
    """This class defines the structure of an action.

    Arguments:
        dataset: Class :class:`~graviti.dataset.dataset.Dataset` instance.
        response: The response of the OpenAPI associated with the action::

                {
                    "id": <str>,
                    "name": <str>,
                    "edition": <int>,
                    "state": <str>,
                    "payload": <str>,
                },

    Attributes:
        name: The name of this action.
        edition: The edition of this action.
        state: The state of this action.
        payload: The payload of this action.

    """

    _repr_attrs = ("edition", "state")

    def __init__(self, dataset: "Dataset", response: Dict[str, Any]) -> None:
        self._dataset = dataset
        self._id: str = response["id"]
        self.name: str = response["name"]
        self.edition: int = response["edition"]
        self.state: str = response["state"]
        self.payload: str = response["payload"]

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.name}")'

    def _edit(
        self,
        *,
        name: Optional[str] = None,
        state: Optional[str] = None,
        payload: Optional[str] = None,
    ) -> None:
        _workspace = self._dataset.workspace
        response = update_action(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            self._dataset.name,
            action=self.name,
            name=name,
            state=state,
            payload=payload,
        )

        self.name = response["name"]
        self.edition = response["edition"]
        self.state = response["state"]
        self.payload = response["payload"]

    @property
    def runs(self) -> "RunManager":
        """Get class :class:`~graviti.manager.action.RunManager` instance.

        Returns:
            Required :class:`~graviti.manager.action.RunManager` instance.

        """
        return RunManager(self)

    def edit(self, *, name: Optional[str] = None, payload: Optional[str] = None) -> None:
        """Update the action.

        Arguments:
            name: The new name of the action.
            payload: The new paylaod of the action.

        """
        self._edit(name=name, payload=payload)

    def enable(self) -> None:
        """Enable the action."""
        self._edit(state="ENABLED")

    def disable(self) -> None:
        """Disable the action."""
        self._edit(state="DISABLED")


class ActionManager:
    """This class defines the operations on the action in Graviti.

    Arguments:
        dataset: :class:`~graviti.manager.dataset.Dataset` instance.

    """

    def __init__(self, dataset: "Dataset") -> None:
        self._dataset = dataset

    def _generate(
        self, offset: int, limit: int, *, sort: SortParam
    ) -> Generator[Action, None, int]:
        _dataset = self._dataset
        _workspace = _dataset.workspace
        response = list_actions(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            sort=sort,
            offset=offset,
            limit=limit,
        )

        for item in response["actions"]:
            yield Action(_dataset, item)

        return response["total_count"]  # type: ignore[no-any-return]

    def create(self, name: str, payload: str) -> Action:
        """Create an action.

        Arguments:
            name: The name of the action.
            payload: The payload of the action.

        Returns:
            The :class:`.Action` instance with the given name.

        """
        _dataset = self._dataset
        _workspace = _dataset.workspace
        response = create_action(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            name=name,
            payload=payload,
        )

        return Action(_dataset, response)

    def get(self, name: str) -> Action:
        """Get the action with the given name.

        Arguments:
            name: The required action name.

        Raises:
            ResourceNameError: When the given name is an empty string.

        Returns:
            The :class:`.Action` instance with the given name.

        """
        check_type("name", name, str)
        if not name:
            raise ResourceNameError("action", name)

        _dataset = self._dataset
        _workspace = _dataset.workspace
        response = get_action(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            action=name,
        )

        return Action(_dataset, response)

    def list(self, *, sort: SortParam = None) -> LazyPagingList[Action]:
        """List the information of actions.

        Arguments:
            sort: The column and the direction the list result sorted by.

        Returns:
            The LazyPagingList of :class:`actions<.Action>` instances.

        """
        return LazyPagingList(
            lambda offset, limit: self._generate(offset, limit, sort=sort),
            LIMIT,
        )

    def delete(self, name: str) -> None:
        """Delete an action.

        Arguments:
            name: The name of the action to be deleted.

        Raises:
            ResourceNameError: When the given name is an empty string.

        """
        check_type("name", name, str)
        if not name:
            raise ResourceNameError("action", name)

        _workspace = self._dataset.workspace
        delete_action(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            self._dataset.name,
            action=name,
        )


class Run(ReprMixin):  # pylint: disable=too-many-instance-attributes
    """This class defines the structure of an action run.

    Arguments:
        action: Class :class:`~graviti.manager.action.Action` instance.
        response: The response of the OpenAPI associated with the action run::

                {
                    "id": <str>,
                    "number": <int>,
                    "name": <str>,
                    "status": <int>,
                    "arguments": <dict>,
                    "started_at": <str>,
                    "ended_at": <str>,
                    "duration": <int>,
                },

    Attributes:
        number: The number of this action run.
        name: The name of this action run.
        status: The status of this action run.
        arguments: The arguments of this action run.
        started_at: The start time of this action run.
        ended_at: The end time of this action run.
        duration: The duration of this action run.

    """

    _repr_attrs = ("status", "arguments", "started_at", "ended_at", "duration")

    def __init__(self, action: Action, response: Dict[str, Any]) -> None:
        self._action = action
        self.number = response["number"]
        self.name = response["name"]
        self.status = response["status"]
        self.arguments: Dict[str, Any] = response["arguments"]
        self.started_at = convert_iso_to_datetime(response["started_at"])
        self.ended_at = convert_iso_to_datetime(response["endeded_at"])
        self.duration = timedelta(seconds=response["duration"])

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.name}")'


class RunManager:
    """This class defines the operations on the action run in Graviti.

    Arguments:
        action: :class:`~graviti.manager.action.Action` instance.

    """

    def __init__(self, action: Action) -> None:
        self._action = action

    def _generate(self, offset: int, limit: int, *, sort: SortParam) -> Generator[Run, None, int]:
        _action = self._action
        _dataset = _action._dataset  # pylint: disable=protected-access
        _workspace = _dataset.workspace
        response = list_action_runs(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            action=_action.name,
            sort=sort,
            offset=offset,
            limit=limit,
        )

        for item in response["runs"]:
            yield Run(_action, item)

        return response["total_count"]  # type: ignore[no-any-return]

    def create(self, arguments: Optional[Dict[str, Any]] = None) -> Run:
        """Run an action manually.

        Arguments:
            arguments: The arguments of the action run.

        Returns:
            The :class:`.Run` instance with the given arguments.

        """
        _action = self._action
        _dataset = _action._dataset  # pylint: disable=protected-access
        _workspace = _dataset.workspace
        response = create_action_run(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            action=_action.name,
            arguments=arguments,
        )

        return Run(_action, response)

    def get(self, run_number: int) -> Run:
        """Get the action with the given name.

        Arguments:
            run_number: The number of the action run.

        Returns:
            The :class:`.Action` instance with the given name.

        """
        check_type("run_number", run_number, int)

        _action = self._action
        _dataset = _action._dataset  # pylint: disable=protected-access
        _workspace = _dataset.workspace
        response = get_action_run(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            action=_action.name,
            run_number=run_number,
        )

        return Run(_action, response)

    def list(self, *, sort: SortParam = None) -> LazyPagingList[Run]:
        """List the information of action runs.

        Arguments:
            sort: The column and the direction the list result sorted by.

        Returns:
            The LazyPagingList of :class:`runs<.Run>` instances.

        """
        return LazyPagingList(
            lambda offset, limit: self._generate(offset, limit, sort=sort),
            LIMIT,
        )



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Tag and TagManager."""

from typing import TYPE_CHECKING, Generator

from graviti.exception import NoCommitsError, ResourceNameError
from graviti.manager.commit import NamedCommit
from graviti.manager.common import CURRENT_COMMIT, LIMIT, check_head_status
from graviti.manager.lazy import LazyPagingList
from graviti.openapi import create_tag, delete_tag, get_tag, list_tags
from graviti.utility import check_type

if TYPE_CHECKING:
    from graviti.manager.dataset import Dataset


class Tag(NamedCommit):
    """This class defines the structure of the tag of a commit.

    Arguments:
        dataset: Class :class:`~graviti.dataset.dataset.Dataset` instance.
        name: The name of the tag.
        commit_id: The commit id.

    """


class TagManager:
    """This class defines the operations on the tag in Graviti.

    Arguments:
        dataset: :class:`~graviti.manager.dataset.Dataset` instance.

    """

    def __init__(self, dataset: "Dataset") -> None:
        self._dataset = dataset

    def _generate(self, offset: int, limit: int) -> Generator[Tag, None, int]:
        _dataset = self._dataset
        _workspace = _dataset.workspace
        response = list_tags(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            offset=offset,
            limit=limit,
        )

        head = _dataset.HEAD
        for item in response["tags"]:
            check_head_status(head, item["name"], item["commit_id"])
            yield Tag.from_response(_dataset, item)

        return response["total_count"]  # type: ignore[no-any-return]

    def create(self, name: str, revision: str = CURRENT_COMMIT) -> Tag:
        """Create a tag for a commit.

        Arguments:
            name: The tag name to be created for the specific commit.
            revision: The information to locate the specific commit, which can be the commit id,
                the branch name, or the tag name. The default value is the current commit of the
                dataset.

        Raises:
            NoCommitsError: When create tags on the default branch without commit history.

        Returns:
            The :class:`.Tag` instance with the given name.

        """
        _dataset = self._dataset
        _workspace = _dataset.workspace
        head = _dataset.HEAD
        if revision is CURRENT_COMMIT:
            _revision = head.commit_id
            if _revision is None:
                raise NoCommitsError(
                    "Creating tags on the default branch without commit history is not allowed."
                    "Please commit a draft first"
                )
        else:
            _revision = revision

        response = create_tag(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            name=name,
            revision=_revision,
        )

        check_head_status(head, _revision, response["commit_id"])
        return Tag.from_response(_dataset, response)

    def get(self, name: str) -> Tag:
        """Get the certain tag with the given name.

        Arguments:
            name: The required tag name.

        Raises:
            ResourceNameError: When the name is an empty string.

        Returns:
            The :class:`.Tag` instance with the given name.

        """
        check_type("name", name, str)
        if not name:
            raise ResourceNameError("tag", name)

        _dataset = self._dataset
        _workspace = _dataset.workspace
        response = get_tag(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            _dataset.name,
            tag=name,
        )
        check_head_status(_dataset.HEAD, name, response["commit_id"])
        return Tag.from_response(_dataset, response)

    def list(self) -> LazyPagingList[Tag]:
        """List the information of tags.

        Returns:
            The LazyPagingList of :class:`tags<.Tag>` instances.

        """
        return LazyPagingList(self._generate, LIMIT)

    def delete(self, name: str) -> None:
        """Delete a tag.

        Arguments:
            name: The tag name to be deleted for the specific commit.

        Raises:
            ResourceNameError: When the name is an empty string.

        """
        check_type("name", name, str)
        if not name:
            raise ResourceNameError("tag", name)

        _workspace = self._dataset.workspace
        delete_tag(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            self._dataset.name,
            tag=name,
        )



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the search."""

from typing import Any, Dict, Optional

from graviti.openapi.requests import open_api_do
from graviti.utility import SortParam


def create_search(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    commit_id: str,
    sheet: str,
    criteria: Dict[str, Any],
    search_id: Optional[str] = None,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `POST /v2/datasets/{workspace}/{dataset}/commits/{commit_id}/sheets\
    /{sheet}/search`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        commit_id: The commit id.
        sheet: The sheet name.
        criteria: The criteria of the search.
        search_id: The search id of this search.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is to get
            all search results.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> create_search(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "BDD100K",
        ...     commit_id = "fde63f357daf46088639e9f57fd81cad",
        ...     sheet = "train",
        ...     criteria = {
        ...         "where": {
        ...            "$or": [
        ...                {
        ...                    "$eq": ["$.filename", "0000f77c-6257be58.jpg"],
        ...                },
        ...                {
        ...                    "$and": [
        ...                        {
        ...                            "$eq": ["$.attribute.weather", "clear"]
        ...                        },
        ...                        {
        ...                            "$eq": ["$.attribute.timeofday", "daytime"]
        ...                        }
        ...                    ]
        ...                }
        ...            ]
        ...        },
        ...        "offset": 0,
        ...        "limit": 128
        ...    }
        ... )
        {
            "data": [
                {
                    "__record_key": "123750493121329585",
                    "filename": "0000f77c-6257be58.jpg",
                    "image": {
                        "url": "https://content-store-prod-vers",
                        "checksum": "dcc197970e607f7576d978972f6fb312911ce005"
                    },
                    "attribute": {
                        "weather": "clear",
                        "scene": "city street",
                        "timeofday": "daytime"
                    }
                },
                ...(total 128 items)
            ],
            "search_id": "5c7de503c88446e8b37a258f71783d7d"
        }

        Use search_id to avoid creating new search.

        >>> create_search(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "BDD100K",
        ...     commit_id = "fde63f357daf46088639e9f57fd81cad",
        ...     sheet = "train",
        ...     criteria = {"where": {"$eq": ["$.filename", "0000f77c-6257be58.jpg"]}}
        ...     search_id = "5c7de503c88446e8b37a258f71783d7d"
        ... )
        {
            "data": [
                {
                    "__record_key": "123750493121329585",
                    "filename": "0000f77c-6257be58.jpg",
                    "image": {
                        "url": "https://content-store-prod-vers",
                        "checksum": "dcc197970e607f7576d978972f6fb312911ce005"
                    },
                    "attribute": {
                        "weather": "clear",
                        "scene": "city street",
                        "timeofday": "daytime"
                    }
                },
                ...(total 128 items)
            ],
            "search_id": "5c7de503c88446e8b37a258f71783d7d"
        }

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/commits/{commit_id}/sheets/{sheet}/search"

    criteria = criteria.copy()
    if offset is not None:
        criteria["offset"] = offset
    if limit is not None:
        criteria["limit"] = limit
    post_data: Dict[str, Any] = {"criteria": criteria}
    if search_id:
        post_data["search_id"] = search_id

    return open_api_do(  # type: ignore[no-any-return]
        "POST", access_key, url, json=post_data
    ).json()


def create_search_history(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    commit_id: Optional[str] = None,
    draft_number: Optional[int] = None,
    sheet: str,
    criteria: Dict[str, Any],
) -> Dict[str, Any]:
    """Execute the OpenAPI `POST /v2/datasets/{workspace}/{dataset}/searches`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The name of the workspace.
        dataset: The name of the dataset.
        commit_id: The commit id.
        draft_number: The draft number.
        sheet: The name of the sheet.
        criteria: The criteria of the search.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> create_search_history(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "portex-test",
        ...     "BDD100K",
        ...     commit_id="89c33716f3834e188ac1ff749c6c270d",
        ...     sheet="train",
        ...     criteria={"where": {"$any_match": ["$.box2ds", {"$eq": ["$.category", 7]}]}},
        ... )
        {
            "id": "53dbbedf35064f21a7b85def60de840e",
            "commit_id": "89c33716f3834e188ac1ff749c6c270d",
            "sheet": "train",
            "criteria": {
                "where": {
                    "$any_match": ["$.box2ds", { "$eq": ["$.category", 7] }]
                }
            },
            "creator": "linjiX",
            "created_at": "2021-03-05T18:58:10Z",
            "total_count": null
        }

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/searches"

    post_data: Dict[str, Any] = {"sheet": sheet, "criteria": criteria}

    if commit_id is not None:
        post_data["commit_id"] = commit_id

    if draft_number is not None:
        post_data["draft_number"] = draft_number

    return open_api_do(  # type: ignore[no-any-return]
        "POST", access_key, url, json=post_data
    ).json()


def list_search_histories(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    commit_id: Optional[str] = None,
    draft_number: Optional[int] = None,
    sheet: Optional[str] = None,
    sort: SortParam = None,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}/searches`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The name of the workspace.
        dataset: The name of the dataset.
        commit_id: The commit id.
        draft_number: The draft number.
        sheet: The name of the sheet.
        sort: The column and the direction the list result sorted by.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is 128.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> list_search_histories(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "portex-test",
        ...     "BDD100K",
        ... )
        {
            "searches": [
                {
                    "id": "a97a31b29c154b82b3049b7f46904b69",
                    "draft_number": 2,
                    "sheet": "test",
                    "criteria": {
                        "where": {
                            "$any_match": ["$.box2ds", { "$eq": ["$.category", 7] }]
                        }
                    },
                    "creator": "linjiX",
                    "created_at": "2021-03-06T18:58:10Z",
                    "total_count": 200
                },
                {
                    "id": "53dbbedf35064f21a7b85def60de840e",
                    "commit_id": "89c33716f3834e188ac1ff749c6c270d",
                    "sheet": "train",
                    "criteria": {
                        "where": {
                            "$any_match": ["$.box2ds", { "$eq": ["$.category", 7] }]
                        }
                    },
                    "creator": "linjiX",
                    "created_at": "2021-03-05T18:58:10Z",
                    "total_count": 1000
                }
            ],
            "offset": 0,
            "record_size": 2,
            "total_count": 2
        }

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/searches"

    params = {
        "commit_id": commit_id,
        "draft_number": draft_number,
        "sheet": sheet,
        "sort": sort,
        "offset": offset,
        "limit": limit,
    }

    return open_api_do("GET", access_key, url, params=params).json()  # type: ignore[no-any-return]


def get_search_history(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    search_id: str,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}/searches/{search_id}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The name of the workspace.
        dataset: The name of the dataset.
        search_id: The search id.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> get_search_history(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "portex-test",
        ...     "BDD100K",
        ...     search_id="53dbbedf35064f21a7b85def60de840e",
        ... )
        {
            "id": "53dbbedf35064f21a7b85def60de840e",
            "commit_id": "89c33716f3834e188ac1ff749c6c270d",
            "sheet": "train",
            "criteria": {
                "where": {
                    "$any_match": ["$.box2ds", { "$eq": ["$.category", 7] }]
                }
            },
            "creator": "linjiX",
            "created_at": "2021-03-05T18:58:10Z",
            "total_count": 1000
        }

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/searches/{search_id}"
    return open_api_do("GET", access_key, url).json()  # type: ignore[no-any-return]


def delete_search_history(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    search_id: str,
) -> None:
    """Execute the OpenAPI `DELETE /v2/datasets/{workspace}/{dataset}/searches/{search_id}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The name of the workspace.
        dataset: The name of the dataset.
        search_id: The search id.

    Examples:
        >>> delete_search_history(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "portex-test",
        ...     "BDD100K",
        ...     search_id="53dbbedf35064f21a7b85def60de840e",
        ... )

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/searches/{search_id}"
    open_api_do("DELETE", access_key, url)


def get_search_total_count(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    search_id: str,
) -> int:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}/searches/{search_id}\
    /total-count`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The name of the workspace.
        dataset: The name of the dataset.
        search_id: The search id.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> get_search_total_count(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "portex-test",
        ...     "BDD100K",
        ...     search_id="53dbbedf35064f21a7b85def60de840e",
        ... )
        1000

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/searches/{search_id}/total-count"

    return open_api_do("GET", access_key, url).json()  # type: ignore[no-any-return]


def list_search_records(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    search_id: str,
    sort: SortParam = None,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}/searches/{search_id}/records`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The name of the workspace.
        dataset: The name of the dataset.
        search_id: The search id.
        sort: The column and the direction the list result sorted by.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is 128.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> list_search_records(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "portex-test",
        ...     "BDD100K",
        ...     search_id="53dbbedf35064f21a7b85def60de840e",
        ... )
        {
            "records": [
                {
                    "__record_key": "2312312312321",
                    "filename": "0000f77c-6257be58.jpg",
                    "image": {
                        "key": "dcc197970e607f7576d978972f6fb312911ce005",
                        "extension": ".jpg",
                        "size": 52344,
                        "width": 800,
                        "heithg": 600
                    },
                    "attribute": {
                        "weather": "clear",
                        "scene": "city street",
                        "timeofday": "daytime"
                    }
                },
                ...(total 5 items)
            ]
        }

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/searches/{search_id}/records"

    params: Dict[str, Any] = {
        "sort": sort,
        "offset": offset,
        "limit": limit,
    }

    return open_api_do("GET", access_key, url, params=params).json()  # type: ignore[no-any-return]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""File module."""

from graviti.file.audio import Audio, RemoteAudio
from graviti.file.base import File, FileBase, RemoteFile
from graviti.file.image import Image, RemoteImage
from graviti.file.point_cloud import PointCloud, RemotePointCloud

__all__ = [
    "Audio",
    "File",
    "FileBase",
    "Image",
    "PointCloud",
    "RemoteAudio",
    "RemoteFile",
    "RemoteImage",
    "RemotePointCloud",
]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Portex record field releated classes."""


from collections import ChainMap
from itertools import chain
from typing import Any
from typing import ChainMap as ChainMapType
from typing import (
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import pyarrow as pa

from graviti.portex.base import PortexType
from graviti.portex.package import Imports
from graviti.utility import INDENT, FrozenNameOrderedDict, NameOrderedDict

_T = TypeVar("_T", bound="Fields")


class FrozenFields(FrozenNameOrderedDict[PortexType]):
    """Represents a frozen fields dict."""

    def __repr__(self) -> str:
        return self._repr1(0)

    def __setitem__(self, key: Union[int, str], value: PortexType) -> None:
        raise TypeError(f"Cannot set item '{key}' in {self.__class__.__name__}")

    def __delitem__(self, key: Union[int, str]) -> None:
        raise TypeError(f"Cannot delete item '{key}' in {self.__class__.__name__}")

    @classmethod
    def _check_value(cls, value: PortexType) -> None:
        if not isinstance(value, PortexType):
            raise TypeError(f'The value in "{cls.__name__}" should be a PortexType')

    def _setitem(self, key: str, value: PortexType) -> None:
        self._check_value(value)
        super()._setitem(key, value)

    def _repr1(self, level: int) -> str:
        lines = ["{"]
        for name, portex_type in self.items():
            lines.append(
                f"{INDENT}'{name}': "  # pylint: disable=protected-access
                f"{portex_type._repr1(level + 1)},"
            )

        lines.append("}")
        return f"\n{level * INDENT}".join(lines)

    def insert(self, index: int, name: str, portex_type: PortexType) -> None:
        """Insert the name and portex_type at the index.

        Arguments:
            index: The index to insert the field.
            name: The name of the field to be inserted.
            portex_type: The portex_type of the field to be inserted.

        Raises:
            TypeError: When calling this method of FrozenFields.

        """
        raise TypeError(f"Cannot insert in {self.__class__.__name__}")

    def astype(self, name: str, portex_type: PortexType) -> None:
        """Convert the type of the field with the given name to the new PortexType.

        Arguments:
            name: The name of the field to convert.
            portex_type: The new PortexType of the field to convert to.

        Raises:
            TypeError: When calling this method of FrozenFields.

        """
        raise TypeError(f"Cannot change the type of '{name}' in {self.__class__.__name__}")

    def rename(self, old_name: str, new_name: str) -> None:
        """Rename the name of a field.

        Arguments:
            old_name: The current name of the field to be renamed.
            new_name: The new name of the field to assign.

        Raises:
            TypeError: When calling this method of FrozenFields.

        """
        raise TypeError(f"Cannot rename '{old_name}' in {self.__class__.__name__}")


class Fields(NameOrderedDict[PortexType], FrozenFields):  # type: ignore[misc]
    """Represents a Portex ``record`` fields dict."""

    def __init__(
        self, fields: Union[Iterable[Tuple[str, PortexType]], Mapping[str, PortexType], None] = None
    ):
        super().__init__(fields)

    @property
    def imports(self) -> Imports:
        """Get the Fields imports.

        Returns:
            The :class:`Imports` instance of this Fields.

        """
        imports = Imports()
        for portex_type in self._data.values():
            imports.update(portex_type.imports)

        return imports

    def insert(self, index: int, name: str, portex_type: PortexType) -> None:
        """Insert the name and portex_type at the index.

        Arguments:
            index: The index to insert the field.
            name: The name of the field to be inserted.
            portex_type: The portex_type of the field to be inserted.

        Raises:
            KeyError: When the name already exists in the Fields.

        """
        self._check_key(name)
        self._check_value(portex_type)

        if self.__contains__(name):
            raise KeyError(f'"{name}" already exists in the Fields')

        self._keys.insert(index, name)
        self._data.__setitem__(name, portex_type)

    def astype(self, name: str, portex_type: PortexType) -> None:
        """Convert the type of the field with the given name to the new PortexType.

        Arguments:
            name: The name of the field to convert.
            portex_type: The new PortexType of the field to convert to.

        Raises:
            KeyError: When the name does not exist in the Fields.

        """
        self._check_value(portex_type)

        if not self.__contains__(name):
            raise KeyError(f'"{name}" does not exist in the Fields')

        self._data.__setitem__(name, portex_type)

    def rename(self, old_name: str, new_name: str) -> None:
        """Rename the name of a field.

        Arguments:
            old_name: The current name of the field to be renamed.
            new_name: The new name of the field to assign.

        """
        self._check_key(new_name)

        self._data.__setitem__(new_name, self._data.pop(old_name))
        self._keys.__setitem__(self._keys.index(old_name), new_name)

    @classmethod
    def from_pyobj(
        cls, content: List[Dict[str, Any]], imports: Optional[Imports] = None
    ) -> "Fields":
        """Create Portex fields dict instance from python list.

        Arguments:
            content: A python list representing a Portex fields dict.
            imports: The imports of the Portex fields dict.

        Returns:
            A Portex fields dict instance created from the input python list.

        """
        return Fields((item["name"], PortexType.from_pyobj(item, imports)) for item in content)

    def to_pyobj(self) -> List[Dict[str, Any]]:
        """Dump the instance to a python list.

        Returns:
            A Python List representation of the fields dict.

        """
        return [{"name": name, **portex_type.to_pyobj(False)} for name, portex_type in self.items()]

    def to_pyarrow(self, *, _to_backend: bool = False) -> List[pa.Field]:
        """Convert the fields to a list of PyArrow Field.

        Returns:
            A list of PyArrow Field representing the fields of Portex record.

        """
        return [
            pa.field(key, value.to_pyarrow(_to_backend=_to_backend)) for key, value in self.items()
        ]

    def copy(self: _T) -> _T:
        """Get a copy of the fields.

        Returns:
            A copy of the fields.

        """
        return self.__class__((name, portex_type.copy()) for name, portex_type in self.items())


UnionFields = Union[FrozenFields, Fields]


class ConnectedFields(MutableMapping[str, PortexType]):
    """Fields composed of FrozenFields and Fields.

    Raises:
        ValueError: When there as repeated field names.

    Arguments:
        multi_fields: The FrozenFields and Fields.

    """

    def __init__(self, multi_fields: Iterable[UnionFields]) -> None:
        multi_fields = list(multi_fields)
        field_keys: Set[str] = set()
        for fields in multi_fields:
            keys = fields.keys()
            repeated_keys = field_keys & keys
            if repeated_keys:
                raise ValueError(f"Repeated field names '{repeated_keys}'")
            field_keys.update(keys)

        self._mapping: ChainMapType[str, PortexType] = ChainMap()
        self._mapping.maps = multi_fields  # type: ignore[assignment]
        self._sequence = multi_fields

    def __len__(self) -> int:
        return sum(map(len, self._sequence))

    def __getitem__(self, key: Union[int, str]) -> PortexType:
        if isinstance(key, int):
            i, j = self._get_indices(key)
            return self._sequence[i][j]

        return self._mapping[key]

    def __setitem__(self, key: Union[int, str], value: PortexType) -> None:
        if isinstance(key, int):
            i, key = self._get_indices(key)
            fields = self._sequence[i]
        else:
            try:
                fields = self._get_fields(key)
            except KeyError:
                fields = self._sequence[-1]

        fields[key] = value

    def __delitem__(self, key: Union[int, str]) -> None:
        if isinstance(key, int):
            i, key = self._get_indices(key)
            fields = self._sequence[i]
        else:
            fields = self._get_fields(key)

        del fields[key]

    def __iter__(self) -> Iterator[str]:
        yield from chain(*self._sequence)

    def _get_fields(self, key: str) -> UnionFields:
        for fields in self._sequence:
            if key in fields:
                return fields

        raise KeyError(key)

    def _get_indices(self, index: int) -> Tuple[int, int]:
        if index < 0:
            index = self.__len__() + index

        offset = index
        for i, length in enumerate(map(len, self._sequence)):
            if offset < length:
                return i, offset
            offset -= length

        raise IndexError("Index out of range")

    def insert(self, index: int, name: str, portex_type: PortexType) -> None:
        """Insert the name and portex_type at the index.

        Arguments:
            index: The index to insert the field.
            name: The name of the field to be inserted.
            portex_type: The portex_type of the field to be inserted.

        Raises:
            ValueError: When the name already exists in the fields.
            TypeError: When trying to insert a field into FrozenFields.

        """
        if name in self._mapping:
            raise ValueError(f"The '{name}' field already exists")

        i, j = self._get_indices(index)

        try:
            self._sequence[i].insert(j, name, portex_type)
            return
        except TypeError:
            pass

        if j == 0 and i > 0:
            fields = self._sequence[i - 1]
            try:
                fields[name] = portex_type
                return
            except TypeError:
                pass

        raise TypeError(f"Cannot insert at index {index} due to frozen fields") from None

    def astype(self, name: str, portex_type: PortexType) -> None:
        """Convert the type of the field with the given name to the new PortexType.

        Arguments:
            name: The name of the field to convert.
            portex_type: The new PortexType of the field to convert to.

        """
        self._get_fields(name).astype(name, portex_type)

    def rename(self, old_name: str, new_name: str) -> None:
        """Rename the name of a field.

        Arguments:
            old_name: The current name of the field to be renamed.
            new_name: The new name of the field to assign.

        """
        self._get_fields(old_name).rename(old_name, new_name)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Package related class."""


from itertools import chain
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Mapping, Tuple, Type, TypeVar

from graviti.utility import AttrDict, ReprMixin, ReprType, UserMapping, urlnorm

if TYPE_CHECKING:
    from graviti.portex.base import PortexType
    from graviti.portex.builtin import PortexBuiltinType
    from graviti.portex.external import PortexExternalType

_T = TypeVar("_T", bound=Type["PortexType"])
_I = TypeVar("_I", bound="Imports")


class Package(AttrDict[_T]):
    """The base class of Portex package."""

    def __setitem__(self, key: str, value: _T) -> None:
        if key in self:
            raise KeyError(f"{key} already exists in package")

        super().__setitem__(key, value)


class BuiltinPackage(Package[Type["PortexBuiltinType"]]):
    """The builtin Portex package used to manage builtin types."""


class LocalPackage(Package[Type["PortexType"]]):
    """The local Portex package used to manage local types."""


class ExternalPackage(Package[Type["PortexExternalType"]]):
    """The external Portex package used to manage external types.

    Arguments:
        url: The git repo url of the external package.
        revision: The git repo revision (tag/commit) of the external package.

    """

    def __init__(self, url: str, revision: str) -> None:
        super().__init__()
        self.url = url
        self.revision = revision

    @property
    def repo(self) -> str:
        """The repo string of the package.

        Returns:
            The "<url>@<rev>" format repo string.

        """
        return f"{self.url}@{self.revision}"


class Subpackage(UserMapping[str, Type["PortexExternalType"]]):
    """The subset of Portex package, used in :class:`Imports`.

    Arguments:
        package: The source package of this subpackage.

    """

    def __init__(self, package: ExternalPackage) -> None:
        self.package = package
        self._data: Dict[str, Type["PortexExternalType"]] = {}

    def __setitem__(self, key: str, value: Type["PortexExternalType"]) -> None:
        type_ = self._data.get(key)
        if type_:
            if type_ != value:
                raise KeyError("Duplicate names")
            return

        for name, type_ in self._data.items():
            if type_ == value:
                if name != key:
                    raise ValueError(
                        "Same type with a different name already exists in the subpackage"
                    )
                return

        self._data[key] = value

    @classmethod
    def from_pyobj(cls, content: Dict[str, Any]) -> "Subpackage":
        """Create :class:`Subpackage` instance from python dict.

        Arguments:
            content: A python dict representing a subpackage.

        Returns:
            A :class:`Subpackage` instance created from the input python dict.

        """
        url, revision = content["repo"].split("@", 1)
        url = urlnorm(url)
        try:
            package = packages.externals[url, revision]
        except KeyError:
            # TODO: Import "build" on toplevel will cause circular imports, to be solved.
            # pylint: disable=import-outside-toplevel
            from graviti.portex.builder import build_package

            package = build_package(url, revision)

        subpackage = cls(package)
        for type_ in content["types"]:
            name = type_["name"]
            subpackage[type_.get("alias", name)] = package[name]

        return subpackage

    def to_pyobj(self) -> Dict[str, Any]:
        """Dump the instance to a python dict.

        Returns:
            A python dict representation of the :class:`Subpackage` instance.

        """
        pyobj: Dict[str, Any] = {"repo": self.package.repo}
        types = []
        for key, value in self._data.items():
            name = value.__name__
            type_ = {"name": name}
            if key != name:
                type_["alias"] = key
            types.append(type_)

        pyobj["types"] = types
        return pyobj


class Packages:
    """The package manager to manage different Portex packages."""

    def __init__(self) -> None:
        self.builtins = BuiltinPackage()
        self.externals: Dict[Tuple[str, str], ExternalPackage] = {}
        self.locals = LocalPackage()


packages = Packages()


class Imports(Mapping[str, Type["PortexType"]], ReprMixin):
    """The imports of the Portex template type.

    Arguments:
        package: The package the portex belongs to.

    """

    _repr_type = ReprType.MAPPING

    def __init__(self) -> None:
        self._subpackages: Dict[str, Subpackage] = {}

    def __len__(self) -> int:
        return sum(map(len, self._subpackages.values()))

    def __getitem__(self, key: str) -> Type["PortexType"]:
        try:
            return packages.builtins[key]
        except KeyError:
            pass

        for subpackage in self._subpackages.values():
            try:
                return subpackage[key]
            except KeyError:
                continue

        raise KeyError(key)

    def __setitem__(self, key: str, value: Type["PortexType"]) -> None:
        type_ = self.get(key)
        if type_:
            if type_ != value:
                raise KeyError("Duplicate names")
            return

        # TODO: Import "PortexExternalType" on toplevel will cause circular imports, to be solved.
        # pylint: disable=import-outside-toplevel
        from graviti.portex.external import PortexExternalType

        if not issubclass(value, PortexExternalType):
            raise TypeError("Local portex type is not supported yet")

        package = value.package
        subpackage = self._subpackages.get(package.repo)
        if not subpackage:
            subpackage = Subpackage(package)
            self._subpackages[package.repo] = subpackage

        subpackage[key] = value

    def __iter__(self) -> Iterator[str]:
        return chain(*self._subpackages.values())

    def update(self, other: "Imports") -> None:
        """Update the imports with another imports.

        Arguments:
            other: An :class:`Imports` instance whose types need to be updated to this imports.

        """
        for key, value in other.items():
            self.__setitem__(key, value)

    @classmethod
    def from_pyobj(cls: Type[_I], content: List[Dict[str, Any]]) -> _I:
        """Create :class:`Imports` instance from python list.

        Arguments:
            content: A python list representing imported types.

        Returns:
            A :class:`Imports` instance created from the input python list.

        """
        imports = cls()
        for pyobj in content:
            subpackage = Subpackage.from_pyobj(pyobj)
            imports.add_subpackage(subpackage)

        return imports

    def to_pyobj(self) -> List[Dict[str, Any]]:
        """Dump the instance to a python list.

        Returns:
            A python list representation of the Portex imported types.

        """
        return [subpackage.to_pyobj() for subpackage in self._subpackages.values()]

    def add_subpackage(self, subpackage: Subpackage) -> None:
        """Add subpackage to this :class:`Imports` instance.

        Arguments:
            subpackage: The subpackage which needs to be added.

        Raises:
            KeyError: When there are duplicate names in the :class:`imports` instance.

        """
        for key in subpackage:
            if self.__contains__(key):
                raise KeyError("Duplicate names")

        self._subpackages[subpackage.package.repo] = subpackage



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Definitions of different operations about the sheet on a draft."""

from typing import TYPE_CHECKING

from graviti.openapi import create_sheet, delete_sheet
from graviti.operation.common import get_schema
from graviti.portex import record

if TYPE_CHECKING:
    from graviti.manager import Dataset


class SheetOperation:
    """This class defines the basic method of the operation about the sheet on a draft.

    Arguments:
        sheet: The sheet name.

    """

    def __init__(self, sheet: str) -> None:
        self.sheet = sheet

    def execute(self, dataset: "Dataset", draft_number: int) -> None:
        """Execute the OpenAPI create sheet.

        Arguments:
            dataset: The Dataset instance.
            draft_number: The draft number.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError


class CreateSheet(SheetOperation):
    """This class defines the operation that create a sheet.

    Arguments:
        sheet: The sheet name.
        schema: The schema of the DataFrame.

    """

    def __init__(self, sheet: str, schema: record) -> None:
        super().__init__(sheet)
        self.schema = schema

    def execute(self, dataset: "Dataset", draft_number: int) -> None:
        """Execute the OpenAPI create sheet.

        Arguments:
            dataset: The Dataset instance.
            draft_number: The draft number.

        """
        portex_schema, avro_schema, arrow_schema = get_schema(self.schema)

        _workspace = dataset.workspace
        create_sheet(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            dataset.name,
            draft_number=draft_number,
            name=self.sheet,
            schema=portex_schema,
            _avro_schema=avro_schema,
            _arrow_schema=arrow_schema,
        )


class DeleteSheet(SheetOperation):
    """This class defines the operation that delete a sheet."""

    def execute(self, dataset: "Dataset", draft_number: int) -> None:
        """Execute the OpenAPI delete sheet.

        Arguments:
            dataset: The Dataset instance.
            draft_number: The draft number.

        """
        _workspace = dataset.workspace
        delete_sheet(
            _workspace.access_key,
            _workspace.url,
            _workspace.name,
            dataset.name,
            draft_number=draft_number,
            sheet=self.sheet,
        )



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
# pylint: disable=c-extension-no-member

"""The Portex builtin types."""


from sys import version_info
from typing import Iterable, Mapping, Optional, Tuple, Type, TypeVar, Union

import pyarrow as pa

from graviti.portex import ptype as PTYPE
from graviti.portex.base import PortexRecordBase, PortexType
from graviti.portex.enum import EnumValues
from graviti.portex.factory import ConnectedFieldsFactory
from graviti.portex.field import Fields
from graviti.portex.package import packages
from graviti.portex.param import Param, Params, param
from graviti.portex.register import PyArrowConversionRegister
from graviti.utility import ModuleMocker

if version_info >= (3, 9):
    # pylint: disable=import-error
    from zoneinfo import ZoneInfo as tz_checker
else:
    try:
        # pylint: disable=import-error
        from pytz import timezone as tz_checker  # type: ignore[import]
    except ModuleNotFoundError:
        tz_checker = ModuleMocker(
            "'pytz' package or 'zoneinfo' module (builtin module after python 3.9) is needed "
            "to support timezone"
        )


builtins = packages.builtins
_E = Union[int, float, str, bool, None]


class PortexBuiltinType(PortexType):  # pylint: disable=abstract-method
    """The base class of Portex builtin type."""

    _T = TypeVar("_T", bound="PortexBuiltinType")

    params = Params()
    packages = builtins

    def __init_subclass__(cls) -> None:
        params = Params(cls.params)
        for name in getattr(cls, "__annotations__", {}):
            parameter = getattr(cls, name, None)
            if isinstance(parameter, tuple):
                params.add(Param(name, *parameter))
                delattr(cls, name)

        cls.params = params

        builtins[cls.__name__] = cls

    def __init__(self, nullable: bool = False) -> None:
        self.nullable = PTYPE.Boolean.check(nullable)

    @classmethod
    def _from_pyarrow(cls: Type[_T], paarray: pa.Array) -> _T:
        return cls()

    def to_builtin(self: _T) -> _T:
        """Expand the top level type to Portex builtin type.

        Returns:
            The expanded Portex builtin type.

        """
        return self


@PyArrowConversionRegister(pa.lib.Type_STRING)
class string(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex primitive type ``string``.

    Arguments:
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = string()
        >>> t
        string()

    """

    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def to_pyarrow(self, *, _to_backend: bool = False) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.string()


@PyArrowConversionRegister(pa.lib.Type_BINARY)
class binary(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex primitive type ``binary``.

    Arguments:
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = binary()
        >>> t
        binary()

    """

    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def to_pyarrow(self, *, _to_backend: bool = False) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.binary()


@PyArrowConversionRegister(pa.lib.Type_BOOL)
class boolean(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex primitive type ``boolean``.

    Arguments:
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = boolean()
        >>> t
        boolean()

    """

    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def to_pyarrow(self, *, _to_backend: bool = False) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.bool_()


@PyArrowConversionRegister(pa.lib.Type_INT32)
class int32(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex primitive type ``int32``.

    Arguments:
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = int32()
        >>> t
        int32()

    """

    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def to_pyarrow(self, *, _to_backend: bool = False) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.int32()


@PyArrowConversionRegister(pa.lib.Type_INT64)
class int64(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex primitive type ``int64``.

    Arguments:
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = int64()
        >>> t
        int64()

    """

    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def to_pyarrow(self, *, _to_backend: bool = False) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.int64()


@PyArrowConversionRegister(pa.lib.Type_FLOAT)
class float32(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex primitive type ``float32``.

    Arguments:
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = float32()
        >>> t
        float32()

    """

    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def to_pyarrow(self, *, _to_backend: bool = False) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.float32()


@PyArrowConversionRegister(pa.lib.Type_DOUBLE)
class float64(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex primitive type ``float64``.

    Arguments:
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = float64()
        >>> t
        float64()

    """

    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def to_pyarrow(self, *, _to_backend: bool = False) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.float64()


@PyArrowConversionRegister(pa.lib.Type_STRUCT)
class record(PortexBuiltinType, PortexRecordBase):  # pylint: disable=invalid-name
    """Portex complex type ``record``.

    Arguments:
        fields: The fields of the record.
        nullable: Whether it is a nullable type.

    Examples:
        Create a record by dict:

        >>> t = record({"f0": int32(), "f1": float32()})
        >>> t
        record(
          fields={
            'f0': int32(),
            'f1': float32(),
          },
        )

        Create a record by tuple list:

        >>> t = record([("f0", string()), ("f1", enum(["v0", "v1"]))])
        >>> t
        record(
          fields={
            'f0': string(),
            'f1': enum(
              values=['v0', 'v1'],
            ),
          },
        )

    """

    _T = TypeVar("_T", bound="record")

    _fields_factory = ConnectedFieldsFactory.from_parameter_name("fields")

    fields: Fields = param(ptype=PTYPE.Fields)
    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def __init__(
        self,
        fields: Union[Iterable[Tuple[str, PortexType]], Mapping[str, PortexType]],
        nullable: bool = False,
    ) -> None:
        self.fields = PTYPE.Fields.check(fields)
        super().__init__(nullable=nullable)

    @classmethod
    def _from_pyarrow(cls: Type[_T], paarray: pa.StructArray) -> _T:
        return cls(
            (field.name, cls.from_pyarrow(paarray.field(field.name))) for field in paarray.type
        )

    def to_pyarrow(self, *, _to_backend: bool = False) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow struct DataType.

        """
        return pa.struct(self.fields.to_pyarrow(_to_backend=_to_backend))


@PyArrowConversionRegister(pa.lib.Type_DICTIONARY)
class enum(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex complex type ``enum``.

    Arguments:
        values: The values of the enum members.
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = enum(["v0", "v1"])
        >>> t
        enum(
          values=['v0', 'v1'],
        )

    """

    _T = TypeVar("_T", bound="enum")

    values: EnumValues = param(ptype=PTYPE.Enum)
    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def __init__(self, values: Iterable[_E], nullable: bool = False) -> None:
        self.values = PTYPE.Enum.check(values)
        super().__init__(nullable=nullable)

    @classmethod
    def _from_pyarrow(cls: Type[_T], paarray: pa.DictionaryArray) -> _T:
        return cls(paarray.dictionary.to_pylist())

    def to_pyarrow(self, *, _to_backend: bool = False) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        min_index, max_index = self.values.index_scope
        if min_index >= -128 and max_index <= 127:
            index_type = pa.int8()

        elif min_index >= -32768 and max_index <= 32767:
            index_type = pa.int16()

        elif min_index >= -2147483648 and max_index <= 2147483647:
            index_type = pa.int32()

        else:
            index_type = pa.int64()

        if _to_backend:
            return index_type

        return pa.dictionary(index_type, self.values.to_pyarrow().type)


@PyArrowConversionRegister(pa.lib.Type_LIST, pa.lib.Type_FIXED_SIZE_LIST)
class array(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex complex type ``array``.

    Arguments:
        items: The item type of the array.
        length: The length of the array.
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = array(int32(0), 100)
        >>> t
        array(
          items=int32(
            minimum=0,
          ),
          length=100,
        )

    """

    _T = TypeVar("_T", bound="array")

    items: PortexType = param(ptype=PTYPE.PortexType)
    length: Optional[int] = param(None, ptype=PTYPE.Integer)
    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def __init__(
        self, items: PortexType, length: Optional[int] = None, nullable: bool = False
    ) -> None:
        self.items = PTYPE.PortexType.check(items)
        self.length = PTYPE.Integer.check(length) if length is not None else None
        super().__init__(nullable=nullable)

    @classmethod
    def _from_pyarrow(cls: Type[_T], paarray: Union[pa.ListArray, pa.FixedSizeListArray]) -> _T:
        patype = paarray.type
        return cls(
            cls.from_pyarrow(paarray.values),
            getattr(patype, "list_size", None),
        )

    def _get_column_count(self) -> int:
        """Get the total column count of the portex type.

        Returns:
            The total column count.

        """
        return self.items._get_column_count()  # pylint: disable=protected-access

    def to_pyarrow(self, *, _to_backend: bool = False) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        list_size = self.length if self.length else -1
        return pa.list_(self.items.to_pyarrow(_to_backend=_to_backend), list_size)


@PyArrowConversionRegister(pa.lib.Type_DATE32)
class date(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex temporal type ``date``.

    Arguments:
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = date()
        >>> t
        date()

    """

    _T = TypeVar("_T", bound="date")

    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def __init__(self, nullable: bool = False) -> None:
        super().__init__(nullable=nullable)

    def to_pyarrow(self, *, _to_backend: bool = False) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.date32()


@PyArrowConversionRegister(pa.lib.Type_TIME32, pa.lib.Type_TIME64)
class time(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex temporal type ``time``.

    Arguments:
        unit: The unit of the time, supports 's', 'ms', 'us' and 'ns'.
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = time("ms")
        >>> t
        times(
          unit='ms',
        )

    """

    _T = TypeVar("_T", bound="time")

    _UNIT_TO_TYPE = {
        "s": pa.time32("s"),
        "ms": pa.time32("ms"),
        "us": pa.time64("us"),
        "ns": pa.time64("ns"),
    }
    _TYPE_TO_UNIT = {value: key for key, value in _UNIT_TO_TYPE.items()}

    unit: str = param(ptype=PTYPE.String, options=["s", "ms", "us", "ns"])
    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def __init__(self, unit: str, nullable: bool = False) -> None:
        self.unit = self.params["unit"].check(unit)
        super().__init__(nullable=nullable)

    @classmethod
    def _from_pyarrow(cls: Type[_T], paarray: Union[pa.Time32Array, pa.Time64Array]) -> _T:
        return cls(cls._TYPE_TO_UNIT[paarray.type])

    def to_pyarrow(self, *, _to_backend: bool = False) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return self._UNIT_TO_TYPE[self.unit]


@PyArrowConversionRegister(pa.lib.Type_TIMESTAMP)
class timestamp(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex temporal type ``timestamp``.

    Arguments:
        unit: The unit of the timestamp, supports 's', 'ms', 'us' and 'ns'.
        tz: The name of the timezone, ``None`` indicates the timestamp is naive.
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = timestamp("ms")
        >>> t
        timestamp(
          unit='ms',
        )
        >>>
        >>> t = timestamp("us", tz="UTC")
        >>> t
        timestamp(
          unit='ms',
          tz='UTC',
        )

    """

    _T = TypeVar("_T", bound="timestamp")

    unit: str = param(ptype=PTYPE.String, options=["s", "ms", "us", "ns"])
    tz: Optional[str] = param(None, ptype=PTYPE.String)
    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def __init__(self, unit: str, tz: Optional[str] = None, nullable: bool = False) -> None:
        self.unit = self.params["unit"].check(unit)
        _tz = PTYPE.String.check(tz) if tz is not None else None
        if _tz is not None:
            tz_checker(_tz)

        self.tz = _tz

        super().__init__(nullable=nullable)

    @classmethod
    def _from_pyarrow(cls: Type[_T], paarray: pa.TimestampArray) -> _T:
        patype = paarray.type
        return cls(patype.unit, patype.tz)

    def to_pyarrow(self, *, _to_backend: bool = False) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.timestamp(self.unit, self.tz)


@PyArrowConversionRegister(pa.lib.Type_DURATION)
class timedelta(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex temporal type ``timedelta``.

    Arguments:
        unit: The unit of the timedelta, supports 's', 'ms', 'us' and 'ns'.
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = timedelta("ms")
        >>> t
        timedelta(
          unit='ms',
        )

    """

    _T = TypeVar("_T", bound="timedelta")

    unit: str = param(ptype=PTYPE.String, options=["s", "ms", "us", "ns"])
    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def __init__(self, unit: str, nullable: bool = False) -> None:
        self.unit = self.params["unit"].check(unit)
        super().__init__(nullable=nullable)

    @classmethod
    def _from_pyarrow(cls: Type[_T], paarray: pa.DurationArray) -> _T:
        return cls(paarray.type.unit)

    def to_pyarrow(self, *, _to_backend: bool = False) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.duration(self.unit)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the dataset."""

from typing import Any, Dict, Optional

from graviti.openapi.requests import open_api_do


def create_dataset(
    access_key: str,
    url: str,
    name: str,
    *,
    alias: str = "",
    storage_config: Optional[str] = None,
    with_draft: Optional[bool] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `POST /v2/datasets`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        name: Name of the dataset, unique for a user.
        alias: Alias of the dataset, default is "".
        storage_config: The auth storage config name.
        with_draft: Whether to create a draft after the dataset is created. The default value of
            this parameter in OpenAPIv2 is False.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> create_dataset(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "MNIST",
        ... )
        {
           "id": "2bc95d506db2401b898067f1045d7f68",
           "name": "OxfordIIITPet",
           "alias": "",
           "workspace": "graviti-example",
           "default_branch": "main",
           "commit_id": None,
           "cover_url": "https://tutu.s3.cn-northwest-1.amazonaws.com.cn/",
           "creator": "czhual",
           "created_at": "2021-03-03T18:58:10Z",
           "updated_at": "2021-03-03T18:58:10Z",
           "is_public": false,
           "storage_config": "exampleConfigName",
           "backend_type": "OSS"
        }

    """
    url = f"{url}/v2/datasets"
    post_data: Dict[str, Any] = {"name": name, "alias": alias}

    if storage_config is not None:
        post_data["storage_config"] = storage_config
    if with_draft is not None:
        post_data["with_draft"] = with_draft

    return open_api_do(  # type: ignore[no-any-return]
        "POST", access_key, url, json=post_data
    ).json()


def get_dataset(access_key: str, url: str, workspace: str, dataset: str) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> get_dataset(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "OxfordIIITPet"
        ... )
        {
           "id": "2bc95d506db2401b898067f1045d7f68",
           "name": "OxfordIIITPet",
           "alias": "Oxford-IIIT Pet",
           "workspace": "graviti-example",
           "default_branch": "main",
           "commit_id": "a0d4065872f245e4ad1d0d1186e3d397",
           "cover_url": "https://tutu.s3.cn-northwest-1.amazonaws.com.cn/",
           "creator": "czhual",
           "created_at": "2021-03-03T18:58:10Z",
           "updated_at": "2021-03-03T18:58:10Z",
           "is_public": false,
           "storage_config": "exampleConfigName",
           "backend_type": "OSS"
        }

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}"
    return open_api_do("GET", access_key, url).json()  # type: ignore[no-any-return]


def list_datasets(
    access_key: str,
    url: str,
    *,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is 128.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> list_datasets("ACCESSKEY-********", "https://api.graviti.com")
        {
           "datasets": [
               {
                   "id": "2bc95d506db2401b898067f1045d7f68",
                   "name": "OxfordIIITPet",
                   "alias": "Oxford-IIIT Pet",
                   "workspace": "graviti-example",
                   "default_branch": "main",
                   "commit_id": "a0d4065872f245e4ad1d0d1186e3d397",
                   "cover_url": "https://tutu.s3.cn-northwest-1.amazonaws.com.cn/",
                   "created_at": "2021-03-03T18:58:10Z",
                   "updated_at": "2021-03-03T18:58:10Z",
                   "creator": "czhual",
                   "is_public": false,
                   "storage_config": "exampleConfigName",
                   "backend_type": "OSS"
               }
            ],
           "offset": 0,
           "record_size": 1,
           "total_count": 1
        }

    """
    url = f"{url}/v2/datasets"

    params = {
        "offset": offset,
        "limit": limit,
    }

    return open_api_do("GET", access_key, url, params=params).json()  # type: ignore[no-any-return]


def update_dataset(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    name: Optional[str] = None,
    alias: Optional[str] = None,
    default_branch: Optional[str] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `PATCH /v2/datasets/{workspace}/{dataset}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        name: New name of the dataset, unique for a user.
        alias: New alias of the dataset.
        default_branch: User's chosen branch.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> update_dataset(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "OxfordIIITPet",
        ...     name="OxfordIIITPets",
        ...     alias="Oxford-IIIT Pet",
        ...     default_branch="main",
        ... )
        {
           "id": "2bc95d506db2401b898067f1045d7f68",
           "name": "OxfordIIITPet",
           "alias": "Oxford-IIIT Pet",
           "workspace": "graviti-example",
           "default_branch": "main",
           "commit_id": "a0d4065872f245e4ad1d0d1186e3d397",
           "cover_url": "https://tutu.s3.cn-northwest-1.amazonaws.com.cn/",
           "created_at": "2021-03-03T18:58:10Z",
           "updated_at": "2021-03-03T18:58:10Z",
           "creator": "czhual",
           "is_public": false,
           "storage_config": "exampleConfigName",
           "backend_type": "OSS"
        }

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}"
    patch_data: Dict[str, Any] = {}

    if name is not None:
        patch_data["name"] = name

    if alias is not None:
        patch_data["alias"] = alias

    if default_branch is not None:
        patch_data["default_branch"] = default_branch

    return open_api_do(  # type: ignore[no-any-return]
        "PATCH", access_key, url, json=patch_data
    ).json()


def delete_dataset(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
) -> None:
    """Execute the OpenAPI `DELETE /v2/datasets/{workspace}/{dataset}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.

    Examples:
        >>> delete_dataset(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "OxfordIIITPet",
        ... )

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}"
    open_api_do("DELETE", access_key, url)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the dataset object."""

from typing import Any, Dict, List, Optional

from graviti.openapi.requests import open_api_do


def get_object_permission(
    access_key: str,
    url: str,
    workspace: str,
    dataset: str,
    *,
    actions: str,
    is_internal: Optional[bool] = None,
    expired: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{workspace}/{dataset}/objects/permissions`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        dataset: Name of the dataset, unique for a user.
        actions: The specific actions including "GET" and "PUT". Supports multiple actions,
            which need to be separated by ``|``, like "GET|PUT".
        is_internal: Whether to return the intranet upload address, the default value in
            the OpenAPI is False.
        expired: Token expiry time in seconds. It cannot be negative.

    Returns:
        The response of OpenAPI.

    Examples:
        Request permission to get dataset data from OSS:

        >>> get_object_permission(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     actions="GET",
        ... )
        {
            "backend_type":"OSS",
            "expire_at":"2022-07-12T06:07:52Z",
            "permission": {
                "AccessKeyId":"LTAI4FjgXD3yFJUaasdasd",
                "AccessKeySecret":"LTAI4FjgXD3yFJJKasdad",
                "SecurityToken":"CAISrgJ1q6Ft5B2yfSjIr5bkKILdaseqw",
                "bucket":"content-store-dev",
                "endpoint":"content-store-dev.oss-cn-qingdao.aliyuncs.com"
            }
        }

        Request permission to put dataset data to OSS:

        >>> get_object_permission(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     actions="PUT",
        ... )
        {
            "backend_type":"OSS",
            "expire_at":"2022-07-12T06:07:52Z",
            "permission": {
                "AccessKeyId":"LTAI4FjgXD3yFJUaasdasd",
                "AccessKeySecret":"LTAI4FjgXD3yFJJKasdad",
                "SecurityToken":"CAISrgJ1q6Ft5B2yfSjIr5bkKILdaseqw",
                "bucket":"content-store-dev",
                "endpoint":"content-store-dev.oss-cn-qingdao.aliyuncs.com",
                "prefix":"051dd0676cc74f548a7e9b7ace45c26b/"
            }
        }

        Request permission to get dataset data from AZURE:

        >>> get_object_permission(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     actions="GET",
        ... )
        {
            "backend_type":"AZURE",
            "expire_at":"2022-07-12T06:07:52Z",
            "permission": {
                "container_name":"graviti210304",
                "account_name":"gra220303",
                "sas_param":"se=2022-07-21T10%3A07Z&sig=*******",
                "endpoint_prefix":"https://gra220303.blob.core.window.net/graviti210304/"
            }
        }

        Request permission to put dataset data to AZURE:

        >>> get_object_permission(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     actions="PUT",
        ... )
        {
            "backend_type":"AZURE",
            "expire_at":"2022-07-12T06:07:52Z",
            "permission": {
                "container_name":"graviti210304",
                "account_name":"gra220303",
                "prefix":"examplePrefix/",
                "sas_param":"se=2022-07-21T10%3A07Z&sig=*******",
                "endpoint_prefix":"https://gra220303.blob.core.window.net/graviti210304/"
            }
        }

        Request permission to get dataset data from S3:

        >>> get_object_permission(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     actions="GET",
        ... )
        {
            "backend_type":"S3",
            "expire_at":"2022-07-12T06:07:52Z",
            "permission": {
                "AccessKeyId":"ASIAQHT******",
                "AccessKeySecret":"Y6x2a2cHIlJdx******",
                "SecurityToken":"FwoGZXIvYXdzEH0aDGYBu******",
                "bucket":"fat-dataplatform",
                "endpoint":"s3.amazonaws.com",
                "region":"us-west-1"
            }
        }

        Request permission to put dataset data to S3:

        >>> get_object_permission(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     actions="PUT",
        ... )
        {
            "backend_type":"S3",
            "expire_at":"2022-07-12T06:07:52Z",
            "permission": {
                "AccessKeyId":"ASIAQHT******",
                "AccessKeySecret":"Y6x2a2cHIlJdx******",
                "SecurityToken":"FwoGZXIvYXdzEH0aDGYBu******",
                "bucket":"fat-dataplatform",
                "endpoint":"s3.amazonaws.com",
                "prefix":"051dd0676cc74f548a7e9b7ace45c26b/",
                "region":"us-west-1"
            }
        }

    """
    url = f"{url}/v2/datasets/{workspace}/{dataset}/objects/permissions"

    params = {
        "actions": actions,
        "is_internal": is_internal,
        "expired": expired,
    }

    return open_api_do("GET", access_key, url, params=params).json()  # type: ignore[no-any-return]


def copy_objects(
    access_key: str,
    url: str,
    workspace: str,
    target_dataset: str,
    *,
    source_dataset: str,
    keys: List[str],
) -> Dict[str, List[str]]:
    """Execute the OpenAPI `POST /v2/datasets/{workspace}/{target_dataset}/objects/copy`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        workspace: The workspace of the dataset.
        target_dataset: The name of the target dataset.
        source_dataset: The name of the source dataset.
        keys: The keys of the objects which need to be copied.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> copy_objects(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     source_dataset="EMINST",
        ...     keys=["xxxxx/xxxxx, xxxxx/xxxxx"]
        ... )
        {
            keys: [
                "yyyyyyy/yyyyyy",
                "yyyyyyy/yyyyyy"
            ]
        }


    """
    url = f"{url}/v2/datasets/{workspace}/{target_dataset}/objects/copy"
    post_data = {"source_dataset": source_dataset, "keys": keys}

    return open_api_do(  # type: ignore[no-any-return]
        "POST", access_key, url, json=post_data
    ).json()



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Functions to get image size."""

import struct
from pathlib import Path
from typing import Tuple

from _io import BufferedReader

from graviti.exception import ImageDecodeError
from graviti.utility import ModuleMocker

try:
    from PIL import Image
except ModuleNotFoundError:
    Image = ModuleMocker(
        "Only support getting the size of "
        "JPEG, PNG, JPEG2000, GIF, BMP, GIF, TIFF, ICO, WebP, Flif formats. "
        "Please install pillow to process this file"
    )


def get_image_size(path: Path) -> Tuple[int, int]:
    """Get the height and width of the input image file.

    Arguments:
        path: The path of the image.

    Returns:
        The height and width of the input image.

    """
    with path.open("rb") as fp:
        header = fp.read(26)
        fp.seek(0)
        for image_format in ImageFormatBase.__subclasses__():
            if image_format.check(header, path.stat().st_size):
                return image_format.get_image_size(header, fp)

    return Image.open(path).size  # type: ignore[no-any-return]


class ImageFormatBase:
    """The base class for different image formats."""

    @classmethod
    def check(cls, header: bytes, size: int) -> bool:
        """Check if the input header fits the current image format.

        Arguments:
            header: The header of the image.
            size: The size of the image.

        Returns:
            Whether if the input header fits the current image format.

        """
        return cls._check(header, size)

    @classmethod
    def get_image_size(cls, header: bytes, fp: BufferedReader) -> Tuple[int, int]:
        """Get the height and width through the input data.

        Arguments:
            header: The header of the image or the entire image.
            fp: The image file pointer.

        Returns:
            The height and width of the image.

        Raises:
            ImageDecodeError: When the input image file is invalid.

        """
        try:
            return cls._get_image_size(header, fp)
        except (struct.error, IndexError, ValueError, TypeError):
            raise ImageDecodeError(f"Invalid {cls.__name__} file") from None

    @staticmethod
    def _get_image_size(header: bytes, fp: BufferedReader) -> Tuple[int, int]:
        raise NotImplementedError

    @staticmethod
    def _check(header: bytes, size: int) -> bool:
        raise NotImplementedError


class JPEG(ImageFormatBase):
    """The class for JPEG image format."""

    @staticmethod
    def _check(header: bytes, size: int) -> bool:
        return size >= 2 and header.startswith(b"\377\330")

    @staticmethod
    def _get_image_size(header: bytes, fp: BufferedReader) -> Tuple[int, int]:
        size = 2
        ftype = 0
        while not 0xC0 <= ftype <= 0xCF or ftype in {0xC4, 0xC8, 0xCC}:
            fp.seek(size, 1)
            byte = fp.read(1)
            while ord(byte) == 0xFF:
                byte = fp.read(1)
            ftype = ord(byte)
            size = struct.unpack(">H", fp.read(2))[0] - 2
        # We are at a SOFn block
        fp.seek(1, 1)  # Skip `precision' byte.
        height, width = struct.unpack(">HH", fp.read(4))
        return height, width


class PNG(ImageFormatBase):
    """The class for PNG image format."""

    @staticmethod
    def _check(header: bytes, size: int) -> bool:
        return size >= 24 and header[1:4] == b"PNG" and header[12:16] == b"IHDR"

    @staticmethod
    def _get_image_size(header: bytes, fp: BufferedReader) -> Tuple[int, int]:
        width, height = struct.unpack(">LL", header[16:24])
        return height, width


class OldPNG(ImageFormatBase):
    """The class for an older version of PNG image format."""

    @staticmethod
    def _check(header: bytes, size: int) -> bool:
        return size >= 16 and header[1:4] == b"PNG"

    @staticmethod
    def _get_image_size(header: bytes, fp: BufferedReader) -> Tuple[int, int]:
        width, height = struct.unpack(">LL", header[8:16])
        return height, width


class GIF(ImageFormatBase):
    """The class for GIF image format."""

    @staticmethod
    def _check(header: bytes, size: int) -> bool:
        return size >= 10 and header[:6] in {b"GIF87a", b"GIF89a"}

    @staticmethod
    def _get_image_size(header: bytes, fp: BufferedReader) -> Tuple[int, int]:
        width, height = struct.unpack("<HH", header[6:10])
        return height, width


class JPEG2000(ImageFormatBase):
    """The class for JPEG 2000 image format."""

    @staticmethod
    def _check(header: bytes, size: int) -> bool:
        return size >= 12 and header.startswith(b"\x00\x00\x00\x0cjP  \r\n\x87\n")

    @staticmethod
    def _get_image_size(header: bytes, fp: BufferedReader) -> Tuple[int, int]:
        fp.seek(48)
        height, width = struct.unpack(">LL", fp.read(8))
        return height, width


class BMP(ImageFormatBase):
    """The class for BMP image format."""

    @staticmethod
    def _check(header: bytes, size: int) -> bool:
        return size >= 26 and header[:2] == b"BM"

    @staticmethod
    def _get_image_size(header: bytes, fp: BufferedReader) -> Tuple[int, int]:
        headersize = struct.unpack("<I", header[14:18])[0]
        if headersize == 12:
            width, height = struct.unpack("<HH", header[18:22])
        elif headersize >= 40:
            width, height = struct.unpack("<ii", header[18:26])
            height = abs(height)  # height is inverted, so abs() the result
        else:
            raise ImageDecodeError("Invalid BMP file")
        return height, width


class TIFF(ImageFormatBase):
    """The class for TIFF image format."""

    _TIFF_TYPES = {
        byte_order_sign: {
            1: (1, byte_order_sign + "B"),  # BYTE
            2: (1, byte_order_sign + "c"),  # ASCII
            3: (2, byte_order_sign + "H"),  # SHORT
            4: (4, byte_order_sign + "L"),  # LONG
            5: (8, byte_order_sign + "LL"),  # RATIONAL
            6: (1, byte_order_sign + "b"),  # SBYTE
            7: (1, byte_order_sign + "c"),  # UNDEFINED
            8: (2, byte_order_sign + "h"),  # SSHORT
            9: (4, byte_order_sign + "l"),  # SLONG
            10: (8, byte_order_sign + "ll"),  # SRATIONAL
            11: (4, byte_order_sign + "f"),  # FLOAT
            12: (8, byte_order_sign + "d"),  # DOUBLE
        }
        for byte_order_sign in (">", "<")
    }

    @staticmethod
    def _check(header: bytes, size: int) -> bool:
        return size >= 8 and header[:4] in {b"II\052\000", b"MM\000\052"}

    @staticmethod
    def _get_image_size(header: bytes, fp: BufferedReader) -> Tuple[int, int]:
        # Standard TIFF, big- or little-endian
        # BigTIFF and other different but TIFF-like formats are not
        # supported currently
        byte_order_sign = ">" if header[:2] == b"MM" else "<"
        # maps TIFF type id to size (in bytes)
        # and python format char for struct
        tiff_types = TIFF._TIFF_TYPES[byte_order_sign]
        ifd_offset = struct.unpack(byte_order_sign + "L", header[4:8])[0]
        fp.seek(ifd_offset)
        # 2 bytes: TagId + 2 bytes: type + 4 bytes: count of values + 4
        # bytes: value offset
        width, height = -1, -1
        digit_type = byte_order_sign + "H"
        for i in range(struct.unpack(digit_type, fp.read(2))[0]):
            entry_offset = ifd_offset + 2 + i * 12
            fp.seek(entry_offset)
            tag = struct.unpack(digit_type, fp.read(2))[0]
            if tag in {256, 257}:
                # if type indicates that value fits into 4 bytes, value
                # offset is not an offset but value itself
                type_ = struct.unpack(digit_type, fp.read(2))[0]
                try:
                    type_size, type_char = tiff_types[type_]
                except KeyError:
                    raise ImageDecodeError(f"Unkown TIFF field type: {type_}") from None
                fp.seek(entry_offset + 8)
                value = int(struct.unpack(type_char, fp.read(type_size))[0])
                if tag == 256:
                    width = value
                else:
                    height = value
            if width > -1 and height > -1:
                break

        if width == -1 or height == -1:
            raise ImageDecodeError(
                "Invalid TIFF file: width and/or height IDS entries are missing."
            )

        return height, width


class ICO(ImageFormatBase):
    """The class for ICO image format."""

    @staticmethod
    def _check(header: bytes, size: int) -> bool:
        return (
            size >= 2
            and struct.unpack("<H", header[:2])[0] == 0
            and struct.unpack("<H", header[2:4])[0] == 1
        )

    @staticmethod
    def _get_image_size(header: bytes, fp: BufferedReader) -> Tuple[int, int]:
        # read the dimensions of the first image
        return header[7], header[6]


class WebP(ImageFormatBase):
    """The class for WebP image format."""

    @staticmethod
    def _check(header: bytes, size: int) -> bool:
        return size >= 24 and header[:4] == b"RIFF" and header[8:12] == b"WEBP"

    @staticmethod
    def _get_image_size(header: bytes, fp: BufferedReader) -> Tuple[int, int]:
        signature = header[12:16]
        if signature == b"VP8L":
            digits = struct.unpack("4B", header[21:25])
            width = 1 + (((digits[1] & 0b111111) << 8) | digits[0])
            height = 1 + (
                ((digits[3] & 0b1111) << 10) | (digits[2] << 2) | ((digits[1] & 0b11000000) >> 6)
            )
        elif signature == b"VP8 ":
            sc_a, sc_b, sc_c = struct.unpack("3B", header[23:26])
            if sc_a != 0x9D or sc_b != 0x01 or sc_c != 0x2A:
                raise ImageDecodeError("Missing start code block for lossy WebP image")
            width, height = struct.unpack("<HH", fp.read(4))
        elif signature == b"VP8X":
            width, height = struct.unpack("<HxH", header[24:] + fp.read(3))
            width, height = width + 1, height + 1
        else:
            raise ImageDecodeError("Invalid WebP file")
        return height, width


class FLIF(ImageFormatBase):
    """The class for Flif image format."""

    @staticmethod
    def _check(header: bytes, size: int) -> bool:
        return size >= 16 and header[:4] == b"FLIF"

    @staticmethod
    def _get_image_size(header: bytes, fp: BufferedReader) -> Tuple[int, int]:
        width, size = FLIF._read_varint(header[6:])
        height, _ = FLIF._read_varint(header[6 + size :])
        return height + 1, width + 1

    @staticmethod
    def _read_varint(data: bytes) -> Tuple[int, int]:
        values = []
        for byte in data:
            value = byte & 0b01111111
            has_leading_bit = byte & 0b10000000
            values.append(value)
            if not has_leading_bit:
                break

        result = 0
        for i, value in enumerate(reversed(values)):
            result |= value << (i * 7)

        return result, len(values)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Graviti indexing related class."""

from typing import TYPE_CHECKING, Any, Iterable, Union, overload

if TYPE_CHECKING:
    from graviti.dataframe.column.series import Series


class ColumnSeriesILocIndexer:
    """Index class for ColumnSeries.iloc."""

    def __init__(self, obj: "Series") -> None:
        self.obj: "Series" = obj

    @overload
    def __getitem__(self, index: int) -> Any:
        ...

    @overload
    def __getitem__(self, index: slice) -> "Series":
        ...

    def __getitem__(self, index: Union[int, slice]) -> Any:
        return self.obj.__getitem__(index)

    @overload
    def __setitem__(self, key: slice, value: Union[Iterable[Any], "Series"]) -> None:
        ...

    @overload
    def __setitem__(self, key: int, value: Any) -> None:
        ...

    def __setitem__(
        self,
        key: Union[int, slice],
        value: Union[Any, Iterable[Any], "Series"],
    ) -> None:
        self.obj.__setitem__(key, value)

    def __delitem__(self, key: Union[int, slice]) -> None:
        self.obj.__delitem__(key)


class ColumnSeriesLocIndexer:
    """Index class for ColumnSeries.loc."""

    def __init__(self, obj: "Series") -> None:
        self.obj: "Series" = obj

    @overload
    def __getitem__(self, index: int) -> Any:
        ...

    @overload
    def __getitem__(self, index: slice) -> "Series":
        ...

    def __getitem__(self, index: Union[int, slice]) -> Any:
        return self.obj.__getitem__(index)

    @overload
    def __setitem__(self, key: slice, value: Union[Iterable[Any], "Series"]) -> None:
        ...

    @overload
    def __setitem__(self, key: int, value: Any) -> None:
        ...

    def __setitem__(
        self,
        key: Union[int, slice],
        value: Union[Any, Iterable[Any], "Series"],
    ) -> None:
        self.obj.__setitem__(key, value)

    def __delitem__(self, key: Union[int, slice]) -> None:
        self.obj.__delitem__(key)
